#include <avr/io.h>
#include "param.h"

.global SIG_OUTPUT_COMPARE0
SIG_OUTPUT_COMPARE0:
	push r16
	in r16, _SFR_IO_ADDR(SREG)
	push r16
	sbi _SFR_IO_ADDR(PORTD), LP
	push YL
	push YH
	push r17
	push r18
	cbi _SFR_IO_ADDR(PORTD), LP

	cbi _SFR_IO_ADDR(PORTD), XSCL
	clr YL
	lds YH, YCnt					// Zeilenadresse laden
	ldi r18, (((XSIZE+XOScan)/4))
	add r18, YL
	inc YH
	cbi _SFR_IO_ADDR(PORTD), FLM

#ifdef LARGEMEM
	in r17, _SFR_IO_ADDR(PORTB)
	push r17
	lds r16, lcd_viewbank
#if LARGEMEM<256
	bst r16, 0 
	bld r17, 0
#elif LARGEMEM<512
	andi r16, 3
	andi r17, 252
	or r17, r16
#elif LARGEMEM<1024
	andi r16, 7
	andi r17, 248
	or r17, r16
#elif LARGEMEM>=1024
	andi r16, 15
	andi r17, 240
	or r17, r16
#endif
	out _SFR_IO_ADDR(PORTB), r17
#endif

Xshift:
	ld r16, Y+
	ld r16, Y+
	ld r16, Y+
	ld r16, Y+
	ld r16, Y+
	ld r16, Y+
	ld r16, Y+
	ld r16, Y+
	cp YL, r18
	brlo Xshift

	cpi YH, YMin
	brne not_firstline
	in r16, _SFR_IO_ADDR(PORTD)
	ldi r17, (1<<MOD)|(1<<FLM)
	eor r16, r17
	out _SFR_IO_ADDR(PORTD), r16
	wdr
not_firstline:
	cpi YH, YScan					// VSync ?
	brlo SkipAdRes
	ldi YH, YMin-1					// Adresszhler Reset
SkipAdRes:
	sts YCnt, YH

#ifdef LARGEMEM
	pop r17
	out _SFR_IO_ADDR(PORTB), r17
#endif

	sbi _SFR_IO_ADDR(PORTD), XSCL

	pop r18
	pop r17
	pop YH
	pop YL
	pop r16
	out _SFR_IO_ADDR(SREG), r16
	pop r16
reti


.global lcd_setpixel
lcd_setpixel:
	ldi r18, hi8(XSIZE)	// r25:24: X
	cpi r24, lo8(XSIZE)	// r23:22: Y
	cpc r25, r18		// r20: color
	brsh exitbit
	ldi r18, hi8(YSIZE)
	cpi r22, lo8(YSIZE)
	cpc r23, r18
	brsh exitbit

	mov r21, r24
	andi r21, 3
	ldi r19, 8
shift:								// r19 = 1<<r21
	cpi r21, 0
	breq exitshift
	lsr r19
	dec r21
	rjmp shift
exitshift:
	ldi r18, hi8(YSIZE/2)
	cpi r22, lo8(YSIZE/2)
	cpc r23, r18
	brlo obere				// Pixel in der oberen oder unteren Displayhlfte ?
	subi r22, lo8(YSIZE/2)
;	sbci r23, hi8(YSIZE/2)	// wird auf jedenfall zu 0, daher egal
	swap r19				// untere Hlfte -> Nibble vertauschen
obere:
	lsr r25
	ror r24
	lsr r25
	ror r24
	mov XL, r24
	mov XH, r22
	subi XH, (-(YMin))
	ld r21, X				// Lade Wert aus SRAM
	tst r20
	brne setpix
	com r19
	and r21, r19
	rjmp clrpix
setpix:
	or r21, r19				// Pixel setzen
clrpix:
	st X, r21				// Wert in SRAM schreiben
normalp:
exitbit:
ret

.global lcd_writebyte		// X: r24		// do not use r25!!!
lcd_writebyte:				// Y: r22:23
	cpi r24, (XSIZE/4)		// DATA: r20
	brsh exitbyte			// nibble: r18
	ldi r21, hi8(YSIZE)
	cpi r22, lo8(YSIZE)
	cpc r23, r21
	brsh exitbyte
	mov XL, r24
	ldi r21, hi8(YSIZE/2)
	cpi r22, lo8(YSIZE/2)
	cpc r23, r21
	brsh untereHaelfte		// Byte in der oberen oder unteren Displayhlfte ?
	mov XH, r22	
	swap r20
	subi XH, (-(YMin))
	ld r21, X
	andi r21, 240			;untere Datenhlfte der SRAM Daten behalten
	mov r19, r20
	andi r19, 15			;erste Hlfte der Daten behalten
	or r21, r19
	st X+,r21
	tst r18				; Nibble Mode?
	brne exitbyte
	ld r21, X
	andi r21, 240			;untere Datenhlfte der SRAM Daten behalten
	swap r20
	andi r20, 15			;zweite Hlfte der Daten behalten
	or r21, r20
	st X+,r21
	ret

untereHaelfte:
	mov XH, r22
	subi XH, lo8(YSIZE/2)
	subi XH, (-(YMin))
	ld r21, X				;Lader Wert aus SRAM
	andi r21, 15			;obere Datenhlfte der SRAM Daten behalten
	mov r19, r20
	andi r19, 240			;erste Hlfte der Daten behalten
	or r21, r19			;Beide Hlften zusammenfgen
	st X+,r21
	tst r18				; Nibble Mode?
	brne exitbyte
	ld r21, X
	andi r21, 15			;obere Datenhlfte der SRAM Daten behalten
	swap r20
	andi r20, 240			;zweite Hlfte der Daten behalten
	or r21, r20
	st X+,r21
exitbyte:
ret

.global lcd_writechar		// X: r24
lcd_writechar:				// Y: r22:23
							// C: r20
	mov ZL, r20
	ldi ZH, TEXT_Y
	mul ZL, ZH
	movw ZL, r0
	subi ZL, lo8(-(font))
	sbci ZH, hi8(-(font))

	ldi r25, TEXT_Y
	clr r18					// write bytes
wchar:
	lpm r20, Z+
	rcall lcd_writebyte
	subi r22, lo8(-1)			// Zeilen Adresse ++
	sbci r23, hi8(-1)
	dec r25
	brne wchar
	clr r1
ret
